import os
from pathlib import Path as _PyPath
from typing import Iterable

from elasticai.creator.hdl.translatable import File, Path


class OnDiskFile(File):
    def __init__(self, full_path: str):
        self._full_path = full_path

    def write_text(self, text: Iterable[str]) -> None:
        full_path = _PyPath(self._full_path)
        folder = full_path.parent
        if not folder.exists():
            os.makedirs(folder)
        with open(full_path, "w") as f:
            f.writelines((f"{line}\n" for line in text))


class OnDiskPath(Path):
    def __init__(self, name: str, parent: str = "."):
        self._full_path = f"{parent}/{name}"

    def create_subpath(self, name: str) -> "Path":
        return OnDiskPath(name, parent=self._full_path)

    def as_file(self, suffix: str) -> "File":
        return OnDiskFile(full_path=f"{self._full_path}{suffix}")



from typing import Iterable, Optional

from elasticai.creator.hdl.translatable import File, Path


class InMemoryFile(File):
    def __init__(self, name):
        self.text: list[str] = []
        self.name = name

    def write_text(self, text: Iterable[str]) -> None:
        for line in text:
            self.text.append(line)


class InMemoryPath(Path):
    def __init__(self, name: str, parent: Optional["InMemoryPath"]):
        self.name = name
        self.children: dict[str, Path | File] = dict()
        self.parent = parent

    def as_file(self, suffix: str) -> InMemoryFile:
        file = InMemoryFile(f"{self.name}{suffix}")
        if len(self.children) > 0:
            raise ValueError(
                "non empty path {}, present children: {}".format(
                    self.name, ", ".join(self.children)
                )
            )
        if self.parent is not None:
            self.parent.children[self.name] = file
        return file

    def __getitem__(self, item):
        return self.children[item]

    def create_subpath(self, subpath_name: str) -> "InMemoryPath":
        subpath = InMemoryPath(name=subpath_name, parent=self)
        self.children[subpath_name] = subpath
        return subpath



from collections import OrderedDict
from typing import cast

import torch

from elasticai.creator.base_modules.hard_sigmoid import HardSigmoid
from elasticai.creator.base_modules.hard_tanh import HardTanh
from elasticai.creator.base_modules.lstm import LSTM
from elasticai.creator.base_modules.lstm_cell import LSTMCell
from elasticai.creator.hdl.design_base.design import Design
from elasticai.creator.nn.fixed_point_arithmetics import FixedPointArithmetics
from elasticai.creator.nn.two_complement_fixed_point_config import FixedPointConfig
from elasticai.creator.nn.vhdl.fp_linear_1d import FPLinear1d
from elasticai.creator.nn.vhdl.fp_linear_1d.design import (
    FPLinear1d as _FPLinear1dDesign,
)
from elasticai.creator.nn.vhdl.lstm.design.fp_lstm_cell import FPLSTMCell
from elasticai.creator.nn.vhdl.lstm.design.lstm import LSTMNetworkDesign
from elasticai.creator.nn.vhdl.module import Module


class LSTMNetwork(torch.nn.Module, Module):
    def __init__(self, layers: list[torch.nn.Module]):
        super().__init__()
        self.lstm = layers[0]
        self.layers = torch.nn.Sequential(
            OrderedDict(
                {f"fp_linear_1d_{i}": layer for i, layer in enumerate(layers[1:])}
            )
        )

    def translate(self) -> LSTMNetworkDesign:
        first_lstm = cast(FixedPointLSTMWithHardActivations, self.lstm)
        total_bits = first_lstm.fixed_point_config.total_bits
        frac_bits = first_lstm.fixed_point_config.frac_bits
        hidden_size = first_lstm.hidden_size
        input_size = first_lstm.input_size
        follow_up_linear_layers = cast(
            list[_FPLinear1dDesign],
            [cast(FPLinear1d, layer).translate() for layer in self.layers],
        )
        return LSTMNetworkDesign(
            lstm=first_lstm.translate(),
            linear_layers=follow_up_linear_layers,
            total_bits=total_bits,
            frac_bits=frac_bits,
            hidden_size=hidden_size,
            input_size=input_size,
        )

    def forward(self, x):
        x, (_, _) = self.lstm(x)
        x = x[:, -1]
        return self.layers(x)


class FixedPointLSTMWithHardActivations(LSTM, Module):
    def __init__(
        self,
        total_bits: int,
        frac_bits: int,
        input_size: int,
        hidden_size: int,
        bias: bool,
    ) -> None:
        config = FixedPointConfig(total_bits=total_bits, frac_bits=frac_bits)

        class LayerFactory:
            def lstm(self, input_size: int, hidden_size: int, bias: bool) -> LSTMCell:
                return LSTMCell(
                    arithmetics=FixedPointArithmetics(config=config),
                    sigmoid_factory=HardSigmoid,
                    tanh_factory=HardTanh,
                    input_size=input_size,
                    hidden_size=hidden_size,
                    bias=bias,
                )

        super().__init__(
            input_size=input_size,
            hidden_size=hidden_size,
            bias=bias,
            batch_first=True,
            layers=LayerFactory(),
        )

        self._config = config

    @property
    def fixed_point_config(self) -> FixedPointConfig:
        return self._config

    def translate(self) -> Design:
        def float_to_signed_int(value: float | list) -> int | list:
            if isinstance(value, list):
                return list(map(float_to_signed_int, value))
            return self._config.as_integer(value)

        cast_weights = lambda x: cast(list[list[list[int]]], x)
        cast_bias = lambda x: cast(list[list[int]], x)

        return FPLSTMCell(
            name="lstm_cell",
            total_bits=self._config.total_bits,
            frac_bits=self._config.frac_bits,
            w_ih=cast_weights(float_to_signed_int(self.cell.linear_ih.weight.tolist())),
            w_hh=cast_weights(float_to_signed_int(self.cell.linear_hh.weight.tolist())),
            b_ih=cast_bias(float_to_signed_int(self.cell.linear_ih.bias.tolist())),
            b_hh=cast_bias(float_to_signed_int(self.cell.linear_hh.bias.tolist())),
            lower_bound_for_hard_sigmoid=cast(int, float_to_signed_int(-3)),
            upper_bound_for_hard_sigmoid=cast(int, float_to_signed_int(3)),
        )



from elasticai.creator.hdl.code_generation.abstract_base_template import (
    TemplateConfig,
    TemplateExpander,
    module_to_package,
)
from elasticai.creator.hdl.design_base.design import Design, Port
from elasticai.creator.hdl.design_base.signal import Signal
from elasticai.creator.hdl.translatable import Path
from elasticai.creator.nn.two_complement_fixed_point_config import FixedPointConfig


class FPHardTanh(Design):
    def __init__(self, total_bits: int, frac_bits: int):
        super().__init__(name="hardtanh")
        self._data_width = total_bits
        fp_config = FixedPointConfig(frac_bits=frac_bits, total_bits=total_bits)
        self._template = TemplateConfig(
            package=module_to_package(self.__module__),
            file_name="fp_hard_tanh.tpl.vhd",
            parameters=dict(
                data_width=str(self._data_width),
                one=str(fp_config.as_integer(1)),
                minus_one=str(fp_config.as_integer(-1)),
            ),
        )

    def save_to(self, destination: "Path"):
        destination.as_file(".vhd").write_text(TemplateExpander(self._template).lines())

    @property
    def port(self) -> Port:
        return Port(
            incoming=[Signal("x", self._data_width)],
            outgoing=[Signal("y", self._data_width)],
        )



from abc import abstractmethod
from typing import Protocol, TypeVar

from elasticai.creator.hdl.translatable import Saveable

T = TypeVar("T", bound=Saveable, covariant=True)


class Tensor(Protocol):
    ...


class Module(Protocol[T]):
    def __call__(self, x) -> Tensor:
        ...

    @abstractmethod
    def translate(self) -> T:
        ...



from abc import ABC, abstractmethod
from dataclasses import dataclass
from itertools import chain, repeat
from string import Template as StringTemplate
from typing import Iterable, Iterator, Mapping, cast

from elasticai.creator.resource_utils import read_text


class AbstractBaseTemplate(ABC):
    """
    `Template` helps you to fill template_resources with values.
    It loads a vhdl template file from the package `elasticai.creator.vhdl.template` and fills template parameters
    with the parameters passed to the constructor.
    Each of these parameters is expected to be either a `str` or a `Iterable[str]`.
    Depending on the type, the parameter will be subject to multi or single line expansion.
    Multi line expansion in the context means the template `"$myValue"` with the parameter `{"myValue": ["1", "2"]}`
    will result in the string `"1\n2"`.
    The **code** method returns the lines of code resulting from the parameters filled into the template.
    """

    def __init__(self, **parameters: str | list[str] | tuple[str]) -> None:
        self._parameters: dict[str, str] = dict()
        self._multiline_parameters: dict[str, Iterable[str]] = dict()
        self.update_parameters(**parameters)
        self._saved_raw_template: list[str] = []

    @abstractmethod
    def _read_raw_template(self) -> Iterator[str]:
        ...

    @property
    def _raw_template(self) -> list[str]:
        if len(self._saved_raw_template) == 0:
            self._saved_raw_template = list(self._read_raw_template())
        return self._saved_raw_template

    def update_parameters(self, **parameters: str | Iterable[str]) -> None:
        (
            single_line_parameters,
            multiline_parameters,
        ) = _split_single_and_multiline_parameters(parameters)
        self._parameters.update(**single_line_parameters)
        self._multiline_parameters.update(**multiline_parameters)

    @property
    def single_line_parameters(self) -> dict[str, str]:
        return dict(**self._parameters)

    @property
    def parameters(self):
        single_items = self._parameters.items()
        multi_items = self._multiline_parameters.items()
        return dict(
            cast(
                Iterator[tuple[str, str] | tuple[str, Iterable[str]]],
                chain(single_items, multi_items),
            )
        )

    @property
    def multi_line_parameters(self):
        return dict(**self._multiline_parameters)

    def lines(self) -> list[str]:
        lines = _expand_template(self._raw_template, **self._parameters)
        lines = _expand_multiline_template(lines, **self._multiline_parameters)
        return list(lines)


def module_to_package(module: str) -> str:
    return ".".join(module.split(".")[:-1])


@dataclass
class TemplateConfig:
    """
    Used in design definition, by the hw designer.
    HW designer just provides template configs and port definitions as a design.

    Contributor of a new translatable module provides design and ml module as well as how to map parameters

    Creator takes these and uses the template expander to generate the correct file
    """

    package: str
    file_name: str
    parameters: dict[str, str | list[str]]


class TemplateExpander:
    """
    Used during translation by the creator tool. HW designer does not need to touch this or inherit from it.
    """

    def _read_raw_template(self) -> Iterator[str]:
        return read_text(
            self.config.package,
            self.config.file_name,
        )

    def __init__(self, config: TemplateConfig):
        super().__init__()
        self.config = config

    def lines(self) -> list[str]:
        single_line_params, multi_line_params = _split_single_and_multiline_parameters(
            self.config.parameters
        )
        template = self._read_raw_template()
        _lines = _expand_template(template, **single_line_params)
        _lines = _expand_multiline_template(_lines, **multi_line_params)
        return list(_lines)


def _split_single_and_multiline_parameters(
    parameters: Mapping[str, str | Iterable[str]],
) -> tuple[dict[str, str], dict[str, Iterable[str]]]:
    single_line_parameters: dict[str, str] = dict(
        cast(
            Iterator[tuple[str, str]],
            filter(lambda i: isinstance(i[1], str), parameters.items()),
        )
    )
    multiline_parameters = dict(
        cast(
            Iterator[tuple[str, Iterable[str]]],
            filter(lambda i: not isinstance(i[1], str), parameters.items()),
        )
    )
    return single_line_parameters, multiline_parameters


def _expand_multiline_template(
    template: str | list[str] | Iterator[str], **kwargs: Iterable[str]
) -> Iterator[str]:
    """Expand a template field to multiple lines, while keeping indentation.
    Example:
        >>> template = "\\t$my_key"
        >>> values = ["hello,", "world", "!"]
        >>> "\\n".join(_expand_multiline_template(template, my_key=values))
        '\\thello,\\n\\tworld\\n\\t!'
    """
    lines = _unify_template_datatype(template)
    for line in lines:
        contains_no_key = True
        for key in kwargs:
            if f"${key}" in line:
                contains_no_key = False
                for placeholder_line, value in zip(repeat(line), kwargs[key]):
                    t = StringTemplate(placeholder_line)
                    yield t.safe_substitute({key: value})
                break
        if contains_no_key:
            yield line


def _unify_template_datatype(
    template: str | list[str] | Iterator[str],
) -> Iterator[str]:
    if hasattr(template, "splitlines") and callable(template.splitlines):
        lines = template.splitlines()
    else:
        lines = template
    yield from lines


def _expand_template(
    template: str | list[str] | Iterator[str], **kwargs: str
) -> Iterator[str]:
    for line in _unify_template_datatype(template):
        yield StringTemplate(line).safe_substitute(kwargs)
